# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Iterable

from .argument import Argument
from .description import Description
from .element import Element
from .printer import Printer


@dataclass(frozen=True)  # type: ignore[misc]
class Method(Element, abc.ABC):
    """Scanner for methods

    Corresponds to event and requests defined on an interface
    """

    name: str
    since: str | None
    description: Description | None
    arg: list[Argument]

    def imports(
        self, interface: str, module_imports: dict[str, str]
    ) -> list[tuple[str, str]]:
        """Get the imports required for each of the interfaces

        :param interface:
            The name of the interface that the method is a part of.
        :param module_imports:
            A mapping from the name of an interface in the associated
            module that the interface comes from.
        :return:
            A list of 2-tuples, each specifying the path to an imported
            module and the imported class.
        """
        current_protocol = module_imports[interface]

        imports = []
        for arg in self.arg:
            if arg.interface is None:
                continue

            if arg.interface == interface:
                continue

            import_class = arg.interface_class
            import_protocol = module_imports[arg.interface]

            if current_protocol == import_protocol:
                import_path = ".{}".format(arg.interface)
            else:
                import_path = "..{}".format(import_protocol)

            imports.append((import_path, import_class))

        return imports

    @property
    @abc.abstractmethod
    def method_type(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def method_args(self) -> Iterable[str]:
        pass

    @abc.abstractmethod
    def output_doc_params(self, printer: Printer) -> None:
        pass

    @abc.abstractmethod
    def output_body(self, printer: Printer, opcode: int) -> None:
        pass

    def output(
        self,
        printer: Printer,
        opcode: int,
        in_class: str,
        module_imports: dict[str, str],
    ) -> None:
        """Generate the output for the given method to the printer"""
        if len(self.arg) > 0:
            printer(f"@{in_class}.{self.method_type}(")
            with printer.indented():
                for arg in self.arg:
                    arg.output(printer)
                if self.since:
                    printer(f"version={self.since},")
            printer(")")
        else:
            if self.since:
                printer(f"@{in_class}.{self.method_type}(version={self.since})")
            else:
                printer(f"@{in_class}.{self.method_type}()")

        # Generate the definition of the method and args
        args = ", ".join(["self"] + list(self.method_args))
        printer("def {}({}):".format(self.name, args))

        with printer.indented():
            # Write the documentation
            self.output_doc(printer)
            # Write out the body of the method
            self.output_body(printer, opcode)

    def output_doc(self, printer: Printer) -> None:
        """Output the documentation for the interface"""
        if self.description:
            self.description.output(printer)
        else:
            printer('"""' + self.name)
        # Parameter and returns documentation
        if self.arg:
            printer()
            self.output_doc_params(printer)
        printer('"""')



# Copyright 2021 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pywayland.utils import AnonymousFile


def test_anonymous_file():
    with AnonymousFile(10) as fd:
        assert fd > 0

    f = AnonymousFile(10)
    f.close()

    f.open()
    with pytest.raises(IOError, match="File is already open"):
        f.open()
    f.close()
    f.close()



# This file has been autogenerated by the pywayland scanner

# Copyright © 2008-2011 Kristian Høgsberg
# Copyright © 2010-2011 Intel Corporation
# Copyright © 2012-2013 Collabora, Ltd.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice (including the
# next paragraph) shall be included in all copies or substantial
# portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import enum

from pywayland.protocol_core import Argument, ArgumentType, Global, Interface, Proxy, Resource
from .wl_keyboard import WlKeyboard
from .wl_pointer import WlPointer
from .wl_touch import WlTouch


class WlSeat(Interface):
    """Group of input devices

    A seat is a group of keyboards, pointer and touch devices. This object is
    published as a global during start up, or when such a device is hot
    plugged.  A seat typically has a pointer and maintains a keyboard focus and
    a pointer focus.
    """

    name = "wl_seat"
    version = 7

    class capability(enum.IntFlag):
        pointer = 1
        keyboard = 2
        touch = 4

    class error(enum.IntEnum):
        missing_capability = 0


class WlSeatProxy(Proxy):
    interface = WlSeat

    @WlSeat.request(
        Argument(ArgumentType.NewId, interface=WlPointer),
    )
    def get_pointer(self):
        """Return pointer object

        The ID provided will be initialized to the
        :class:`~pywayland.protocol.wayland.WlPointer` interface for this seat.

        This request only takes effect if the seat has the pointer capability,
        or has had the pointer capability in the past. It is a protocol
        violation to issue this request on a seat that has never had the
        pointer capability. The missing_capability error will be sent in this
        case.

        :returns:
            :class:`~pywayland.protocol.wayland.WlPointer` -- seat pointer
        """
        id = self._marshal_constructor(0, WlPointer)
        return id

    @WlSeat.request(
        Argument(ArgumentType.NewId, interface=WlKeyboard),
    )
    def get_keyboard(self):
        """Return keyboard object

        The ID provided will be initialized to the
        :class:`~pywayland.protocol.wayland.WlKeyboard` interface for this
        seat.

        This request only takes effect if the seat has the keyboard capability,
        or has had the keyboard capability in the past. It is a protocol
        violation to issue this request on a seat that has never had the
        keyboard capability. The missing_capability error will be sent in this
        case.

        :returns:
            :class:`~pywayland.protocol.wayland.WlKeyboard` -- seat keyboard
        """
        id = self._marshal_constructor(1, WlKeyboard)
        return id

    @WlSeat.request(
        Argument(ArgumentType.NewId, interface=WlTouch),
    )
    def get_touch(self):
        """Return touch object

        The ID provided will be initialized to the
        :class:`~pywayland.protocol.wayland.WlTouch` interface for this seat.

        This request only takes effect if the seat has the touch capability, or
        has had the touch capability in the past. It is a protocol violation to
        issue this request on a seat that has never had the touch capability.
        The missing_capability error will be sent in this case.

        :returns:
            :class:`~pywayland.protocol.wayland.WlTouch` -- seat touch
            interface
        """
        id = self._marshal_constructor(2, WlTouch)
        return id

    @WlSeat.request(version=5)
    def release(self):
        """Release the seat object

        Using this request a client can tell the server that it is not going to
        use the seat object anymore.
        """
        self._marshal(3)
        self._destroy()


class WlSeatResource(Resource):
    interface = WlSeat

    @WlSeat.event(
        Argument(ArgumentType.Uint),
    )
    def capabilities(self, capabilities):
        """Seat capabilities changed

        This is emitted whenever a seat gains or loses the pointer, keyboard or
        touch capabilities.  The argument is a capability enum containing the
        complete set of capabilities this seat has.

        When the pointer capability is added, a client may create a
        :class:`~pywayland.protocol.wayland.WlPointer` object using the
        :func:`WlSeat.get_pointer()` request. This object will receive pointer
        events until the capability is removed in the future.

        When the pointer capability is removed, a client should destroy the
        :class:`~pywayland.protocol.wayland.WlPointer` objects associated with
        the seat where the capability was removed, using the
        :func:`WlPointer.release()
        <pywayland.protocol.wayland.WlPointer.release>` request. No further
        pointer events will be received on these objects.

        In some compositors, if a seat regains the pointer capability and a
        client has a previously obtained
        :class:`~pywayland.protocol.wayland.WlPointer` object of version 4 or
        less, that object may start sending pointer events again. This behavior
        is considered a misinterpretation of the intended behavior and must not
        be relied upon by the client.
        :class:`~pywayland.protocol.wayland.WlPointer` objects of version 5 or
        later must not send events if created before the most recent event
        notifying the client of an added pointer capability.

        The above behavior also applies to
        :class:`~pywayland.protocol.wayland.WlKeyboard` and
        :class:`~pywayland.protocol.wayland.WlTouch` with the keyboard and
        touch capabilities, respectively.

        :param capabilities:
            capabilities of the seat
        :type capabilities:
            `ArgumentType.Uint`
        """
        self._post_event(0, capabilities)

    @WlSeat.event(
        Argument(ArgumentType.String),
        version=2,
    )
    def name(self, name):
        """Unique identifier for this seat

        In a multi-seat configuration the seat name can be used by clients to
        help identify which physical devices the seat represents.

        The seat name is a UTF-8 string with no convention defined for its
        contents. Each name is unique among all :class:`WlSeat` globals. The
        name is only guaranteed to be unique for the current compositor
        instance.

        The same seat names are used for all clients. Thus, the name can be
        shared across processes to refer to a specific :class:`WlSeat` global.

        The name event is sent after binding to the seat global. This event is
        only sent once per seat object, and the name does not change over the
        lifetime of the :class:`WlSeat` global.

        Compositors may re-use the same seat name if the :class:`WlSeat` global
        is destroyed and re-created later.

        :param name:
            seat identifier
        :type name:
            `ArgumentType.String`
        """
        self._post_event(1, name)


class WlSeatGlobal(Global):
    interface = WlSeat


WlSeat._gen_c()
WlSeat.proxy_class = WlSeatProxy
WlSeat.resource_class = WlSeatResource
WlSeat.global_class = WlSeatGlobal



# This file has been autogenerated by the pywayland scanner

# Copyright © 2014, 2015 Collabora, Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import enum

from pywayland.protocol_core import Argument, ArgumentType, Global, Interface, Proxy, Resource


class ZwpLinuxDmabufFeedbackV1(Interface):
    """Dmabuf feedback

    This object advertises dmabuf parameters feedback. This includes the
    preferred devices and the supported formats/modifiers.

    The parameters are sent once when this object is created and whenever they
    change. The done event is always sent once after all parameters have been
    sent. When a single parameter changes, all parameters are re-sent by the
    compositor.

    Compositors can re-send the parameters when the current client buffer
    allocations are sub-optimal. Compositors should not re-send the parameters
    if re-allocating the buffers would not result in a more optimal
    configuration. In particular, compositors should avoid sending the exact
    same parameters multiple times in a row.

    The tranche_target_device and tranche_modifier events are grouped by
    tranches of preference. For each tranche, a tranche_target_device, one
    tranche_flags and one or more tranche_modifier events are sent, followed by
    a tranche_done event finishing the list. The tranches are sent in
    descending order of preference. All formats and modifiers in the same
    tranche have the same preference.

    To send parameters, the compositor sends one main_device event, tranches
    (each consisting of one tranche_target_device event, one tranche_flags
    event, tranche_modifier events and then a tranche_done event), then one
    done event.
    """

    name = "zwp_linux_dmabuf_feedback_v1"
    version = 4

    class tranche_flags(enum.IntFlag):
        scanout = 1


class ZwpLinuxDmabufFeedbackV1Proxy(Proxy):
    interface = ZwpLinuxDmabufFeedbackV1

    @ZwpLinuxDmabufFeedbackV1.request()
    def destroy(self):
        """Destroy the feedback object

        Using this request a client can tell the server that it is not going to
        use the wp_linux_dmabuf_feedback object anymore.
        """
        self._marshal(0)
        self._destroy()


class ZwpLinuxDmabufFeedbackV1Resource(Resource):
    interface = ZwpLinuxDmabufFeedbackV1

    @ZwpLinuxDmabufFeedbackV1.event()
    def done(self):
        """All feedback has been sent

        This event is sent after all parameters of a wp_linux_dmabuf_feedback
        object have been sent.

        This allows changes to the wp_linux_dmabuf_feedback parameters to be
        seen as atomic, even if they happen via multiple events.
        """
        self._post_event(0)

    @ZwpLinuxDmabufFeedbackV1.event(
        Argument(ArgumentType.FileDescriptor),
        Argument(ArgumentType.Uint),
    )
    def format_table(self, fd, size):
        """Format and modifier table

        This event provides a file descriptor which can be memory-mapped to
        access the format and modifier table.

        The table contains a tightly packed array of consecutive format +
        modifier pairs. Each pair is 16 bytes wide. It contains a format as a
        32-bit unsigned integer, followed by 4 bytes of unused padding, and a
        modifier as a 64-bit unsigned integer. The native endianness is used.

        The client must map the file descriptor in read-only private mode.

        Compositors are not allowed to mutate the table file contents once this
        event has been sent. Instead, compositors must create a new, separate
        table file and re-send feedback parameters. Compositors are allowed to
        store duplicate format + modifier pairs in the table.

        :param fd:
            table file descriptor
        :type fd:
            `ArgumentType.FileDescriptor`
        :param size:
            table size, in bytes
        :type size:
            `ArgumentType.Uint`
        """
        self._post_event(1, fd, size)

    @ZwpLinuxDmabufFeedbackV1.event(
        Argument(ArgumentType.Array),
    )
    def main_device(self, device):
        """Preferred main device

        This event advertises the main device that the server prefers to use
        when direct scan-out to the target device isn't possible. The
        advertised main device may be different for each
        wp_linux_dmabuf_feedback object, and may change over time.

        There is exactly one main device. The compositor must send at least one
        preference tranche with tranche_target_device equal to main_device.

        Clients need to create buffers that the main device can import and read
        from, otherwise creating the dmabuf
        :class:`~pywayland.protocol.wayland.WlBuffer` will fail (see the
        wp_linux_buffer_params.create and create_immed requests for details).
        The main device will also likely be kept active by the compositor, so
        clients can use it instead of waking up another device for power
        savings.

        In general the device is a DRM node. The DRM node type (primary vs.
        render) is unspecified. Clients must not rely on the compositor sending
        a particular node type. Clients cannot check two devices for equality
        by comparing the dev_t value.

        If explicit modifiers are not supported and the client performs buffer
        allocations on a different device than the main device, then the client
        must force the buffer to have a linear layout.

        :param device:
            device dev_t value
        :type device:
            `ArgumentType.Array`
        """
        self._post_event(2, device)

    @ZwpLinuxDmabufFeedbackV1.event()
    def tranche_done(self):
        """A preference tranche has been sent

        This event splits tranche_target_device and tranche_modifier events in
        preference tranches. It is sent after a set of tranche_target_device
        and tranche_modifier events; it represents the end of a tranche. The
        next tranche will have a lower preference.
        """
        self._post_event(3)

    @ZwpLinuxDmabufFeedbackV1.event(
        Argument(ArgumentType.Array),
    )
    def tranche_target_device(self, device):
        """Target device

        This event advertises the target device that the server prefers to use
        for a buffer created given this tranche. The advertised target device
        may be different for each preference tranche, and may change over time.

        There is exactly one target device per tranche.

        The target device may be a scan-out device, for example if the
        compositor prefers to directly scan-out a buffer created given this
        tranche. The target device may be a rendering device, for example if
        the compositor prefers to texture from said buffer.

        The client can use this hint to allocate the buffer in a way that makes
        it accessible from the target device, ideally directly. The buffer must
        still be accessible from the main device, either through direct import
        or through a potentially more expensive fallback path. If the buffer
        can't be directly imported from the main device then clients must be
        prepared for the compositor changing the tranche priority or making
        :class:`~pywayland.protocol.wayland.WlBuffer` creation fail (see the
        wp_linux_buffer_params.create and create_immed requests for details).

        If the device is a DRM node, the DRM node type (primary vs. render) is
        unspecified. Clients must not rely on the compositor sending a
        particular node type. Clients cannot check two devices for equality by
        comparing the dev_t value.

        This event is tied to a preference tranche, see the tranche_done event.

        :param device:
            device dev_t value
        :type device:
            `ArgumentType.Array`
        """
        self._post_event(4, device)

    @ZwpLinuxDmabufFeedbackV1.event(
        Argument(ArgumentType.Array),
    )
    def tranche_formats(self, indices):
        """Supported buffer format modifier

        This event advertises the format + modifier combinations that the
        compositor supports.

        It carries an array of indices, each referring to a format + modifier
        pair in the last received format table (see the format_table event).
        Each index is a 16-bit unsigned integer in native endianness.

        For legacy support, DRM_FORMAT_MOD_INVALID is an allowed modifier. It
        indicates that the server can support the format with an implicit
        modifier. When a buffer has DRM_FORMAT_MOD_INVALID as its modifier, it
        is as if no explicit modifier is specified. The effective modifier will
        be derived from the dmabuf.

        A compositor that sends valid modifiers and DRM_FORMAT_MOD_INVALID for
        a given format supports both explicit modifiers and implicit modifiers.

        Compositors must not send duplicate format + modifier pairs within the
        same tranche or across two different tranches with the same target
        device and flags.

        This event is tied to a preference tranche, see the tranche_done event.

        For the definition of the format and modifier codes, see the
        wp_linux_buffer_params.create request.

        :param indices:
            array of 16-bit indexes
        :type indices:
            `ArgumentType.Array`
        """
        self._post_event(5, indices)

    @ZwpLinuxDmabufFeedbackV1.event(
        Argument(ArgumentType.Uint),
    )
    def tranche_flags(self, flags):
        """Tranche flags

        This event sets tranche-specific flags.

        The scanout flag is a hint that direct scan-out may be attempted by the
        compositor on the target device if the client appropriately allocates a
        buffer. How to allocate a buffer that can be scanned out on the target
        device is implementation-defined.

        This event is tied to a preference tranche, see the tranche_done event.

        :param flags:
            tranche flags
        :type flags:
            `ArgumentType.Uint`
        """
        self._post_event(6, flags)


class ZwpLinuxDmabufFeedbackV1Global(Global):
    interface = ZwpLinuxDmabufFeedbackV1


ZwpLinuxDmabufFeedbackV1._gen_c()
ZwpLinuxDmabufFeedbackV1.proxy_class = ZwpLinuxDmabufFeedbackV1Proxy
ZwpLinuxDmabufFeedbackV1.resource_class = ZwpLinuxDmabufFeedbackV1Resource
ZwpLinuxDmabufFeedbackV1.global_class = ZwpLinuxDmabufFeedbackV1Global



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
import xml.etree.ElementTree as ET

from .description import Description
from .entry import Entry
from .element import Element
from .printer import Printer


@dataclass(frozen=True)
class Enum(Element):
    """Scanner for enum objects

    Required attributes: `name` and `since`

    Child elements: `description` and `entry`
    """

    name: str
    since: str | None
    is_bitfield: bool
    description: Description | None
    entry: list[Entry]

    @classmethod
    def parse(cls, element: ET.Element) -> Enum:
        is_bitfield = cls.parse_optional_attribute(element, "bitfield") == "true"
        return Enum(
            name=cls.parse_attribute(element, "name"),
            since=cls.parse_optional_attribute(element, "since"),
            is_bitfield=is_bitfield,
            description=cls.parse_optional_child(element, Description, "description"),
            entry=cls.parse_repeated_child(element, Entry, "entry"),
        )

    def output(self, printer: Printer) -> None:
        """Generate the output for the enum to the printer"""
        name = self.name if self.name != "version" else "version_"
        if self.is_bitfield:
            printer("class {0}(enum.IntFlag):".format(name))
        else:
            printer("class {0}(enum.IntEnum):".format(name))
        with printer.indented():
            for entry in self.entry:
                entry.output(self.name, printer)



# Copyright 2015 Sean Vig
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
import enum
import xml.etree.ElementTree as ET

from .description import Description
from .element import Element
from .printer import Printer

NO_IFACE_NAME = "interface"


@enum.unique
class ArgumentType(enum.Enum):
    Int = enum.auto()
    Uint = enum.auto()
    Fixed = enum.auto()
    String = enum.auto()
    Object = enum.auto()
    NewId = enum.auto()
    Array = enum.auto()
    FileDescriptor = enum.auto()


@dataclass(frozen=True)
class Argument(Element):
    """Argument to a request or event method

    Required attributes: `name` and `type`

    Optional attributes: `summary`, `interface`, and `allow-null`

    Child elements: `description`
    """

    name: str
    type: ArgumentType
    summary: str | None
    interface: str | None
    allow_null: bool
    enum: str | None
    description: Description | None

    @classmethod
    def parse(cls, element: ET.Element) -> Argument:
        arg_type_str = cls.parse_attribute(element, "type")
        if arg_type_str == "int":
            argument_type = ArgumentType.Int
        elif arg_type_str == "uint":
            argument_type = ArgumentType.Uint
        elif arg_type_str == "fixed":
            argument_type = ArgumentType.Fixed
        elif arg_type_str == "string":
            argument_type = ArgumentType.String
        elif arg_type_str == "object":
            argument_type = ArgumentType.Object
        elif arg_type_str == "new_id":
            argument_type = ArgumentType.NewId
        elif arg_type_str == "array":
            argument_type = ArgumentType.Array
        elif arg_type_str == "fd":
            argument_type = ArgumentType.FileDescriptor
        else:
            raise ValueError(f"Invalid argument type: {argument_type}")

        allow_null = cls.parse_optional_attribute(element, "allow-null") == "true"
        return cls(
            name=cls.parse_attribute(element, "name"),
            type=argument_type,
            summary=cls.parse_optional_attribute(element, "summary"),
            interface=cls.parse_optional_attribute(element, "interface"),
            allow_null=allow_null,
            enum=cls.parse_optional_attribute(element, "enum"),
            description=cls.parse_optional_child(element, Description, "description"),
        )

    @property
    def interface_class(self) -> str:
        """Returns the Interface class name

        Gives the class name for the Interface coresponding to the type of the
        argument.
        """
        assert self.interface is not None
        return "".join(x.capitalize() for x in self.interface.split("_"))

    def output(self, printer: Printer) -> None:
        args = [f"ArgumentType.{self.type.name}"]
        if self.interface is not None:
            args.append(f"interface={self.interface_class}")
        if self.allow_null:
            args.append("nullable=True")

        printer(f"Argument({', '.join(args)}),")

    def output_doc_param(self, printer: Printer) -> None:
        """Document the argument as a parameter"""
        # Output the parameter and summary
        printer.doc(":param {}:".format(self.name))
        if self.summary:
            with printer.indented():
                printer.docstring(self.summary)

        # Determine the type to be output
        if self.interface:
            arg_type = self.interface
        else:
            arg_type = "`{}`".format(self.type)

        # Output the parameter type
        printer.doc(f":type {self.name}:")
        with printer.indented():
            if self.allow_null:
                printer.doc(f"{arg_type} or `None`")
            else:
                printer.doc(f"{arg_type}")

    def output_doc_ret(self, printer: Printer) -> None:
        """Document the argument as a return"""
        # Determine the type to be output
        if self.interface:
            arg_type = self.interface
        else:
            # Only new_id's are returned, the only corner case here is for
            # wl_registry.bind, so no interface => Proxy
            arg_type = ":class:`pywayland.client.proxy.Proxy` of specified Interface"

        # Output the type and summary
        printer.doc(":returns:")
        with printer.indented():
            if self.summary:
                printer.docstring("{} -- {}".format(arg_type, self.summary))
            else:
                printer.docstring("{}".format(arg_type))
